// Trifield LR→LRC Decoder component for foobar2000 v1.4+
// Copyright (c) 2017-2018 Janne Hyvärinen
//
// Converted to a component from trifield23.hh:
/* ------------------------------------------------------------
trifield23.hh:
author: "AmbisonicDecoderToolkit"
copyright: "(c) Aaron J. Heller 2013"
license: "GPL"
name: "Trifield LR->LRC decoder"
version: "1.0"
Code generated with Faust 2.0.a43 (http://faust.grame.fr)
------------------------------------------------------------ */

#define VERSION "1.1.2"

// Changes:
// 1.1.2 (2018-08-27): Fixed left and right channels being reversed
// 1.1.1 (2018-01-06): No changes in functionality. Dynamic C runtime version for foobar2000 v1.4
// 1.1   (2017-06-07): Added more information to the about dialog. Lowered minimum crossover frequency to 350 Hz. Removed 192 kHz upper limit for sampling frequency
// 1.0   (2017-06-02): Initial version

#include "../foobar2000/ATLHelpers/ATLHelpersLean.h"
#include "../foobar2000/SDK/foobar2000.h"
#include "resource.h"
#include <math.h>

DECLARE_COMPONENT_VERSION("Trifield LR→LRC Decoder", VERSION,
  "An implementation of Michael Gerzon's Trifield decoder.\n"
  "\n"
  "The Trifield process is a form of audio rendering in which a conventional two-channel stereo signal is decoded "
  "(using Ambisonic principles) to an additional number of loudspeakers, typically three in the form of a Left-Center-Right "
  "front stage. The technique provides significant improvement to image stability, especially when the listener is moving "
  "or off-axis. It also addresses the problem in conventional stereo, where center images are recessed and less distinct "
  "due to the comb filtering arising from two sound sources emitting correlated signals.\n"
  "\n"
  "Original software implementation (c) 2013 Aaron J. Heller.\n"
  "foobar2000 version (c) 2017-2018 Janne Hyvärinen.\n"
  "Licensed under GNU Affero General Public License.");

VALIDATE_COMPONENT_FILENAME("foo_dsp_trifield.dll");

static void RunDSPConfigPopup(const dsp_preset &p_data, HWND p_parent, dsp_preset_edit_callback &p_callback);

#define default_strength    1000
#ifndef FAUSTFLOAT
#define FAUSTFLOAT float
#endif  

class trifield_dsp : public dsp_impl_base {
private:
    unsigned int fSamplingFreq;
    FAUSTFLOAT cdx;
    FAUSTFLOAT phihf;
    FAUSTFLOAT philf;
    FAUSTFLOAT width;
    unsigned int xover;
    float fVec0[1024], fRec0[2], fRec1[3], fRec2[2], fRec3[2], fRec4[3], fConst1, fConst2;
    unsigned int IOTA;

    void init(unsigned int samplingFreq)
    {
        fSamplingFreq = samplingFreq;
        unsigned int iConst0 = (fSamplingFreq > 0) ? fSamplingFreq : 1;
        fConst1 = (3.1415927f / (float)iConst0);
        fConst2 = (0.0029136287f * (float)iConst0);
        IOTA = 0;
        for (int i = 0; i < 1024; i++) fVec0[i] = 0.0f;
        for (int i = 0; i < 2; i++) fRec0[i] = 0.0f;
        for (int i = 0; i < 3; i++) fRec1[i] = 0.0f;
        for (int i = 0; i < 2; i++) fRec2[i] = 0.0f;
        for (int i = 0; i < 2; i++) fRec3[i] = 0.0f;
        for (int i = 0; i < 3; i++) fRec4[i] = 0.0f;
    }

    inline float faustpower2_f(float value) { return (value * value); }

    inline void compute(unsigned int count, const audio_sample *input, audio_sample *output)
    {
        float fSlow0 = tanf((fConst1 * (float)xover));
        float fSlow1 = faustpower2_f(fSlow0);
        float fSlow2 = (1.0f + (fSlow0 * (2.0f + fSlow0)));
        float fSlow3 = (fSlow1 / fSlow2);
        float fSlow4 = (1.7453292e-05f * (float)philf);
        float fSlow5 = (1.0f / fSlow2);
        float fSlow6 = (2.0f * (fSlow1 - 1.0f));
        float fSlow7 = (1.0f + (fSlow0 * (fSlow0 - 2.0f)));
        float fSlow8 = (1.7453292e-05f * (float)phihf);
        float fSlow9 = (0.0f - (2.0f / fSlow2));
        float fSlow10 = (0.001f * (float)width);
        unsigned int iSlow11 = (unsigned int)(0.5f + (fConst2 * (float)cdx));

        for (unsigned int i = 0; i < count; i++) {
            fRec0[0] = (fSlow4 + (0.999f * fRec0[1]));
            float fTemp0 = (float)*input++;
            float fTemp1 = (float)*input++;
            fRec1[0] = ((0.70710677f * (fTemp0 + fTemp1)) - (fSlow5 * ((fSlow6 * fRec1[1]) + (fSlow7 * fRec1[2]))));
            float fTemp2 = (fRec1[2] + (fRec1[0] + (2.0f * fRec1[1])));
            fRec2[0] = (fSlow8 + (0.999f * fRec2[1]));
            float fTemp3 = (0.0f - ((fSlow9 * fRec1[1]) + (fSlow5 * (fRec1[0] + fRec1[2]))));
            float fTemp4 = ((fSlow3 * (sinf(fRec0[0]) * fTemp2)) + (sinf(fRec2[0]) * fTemp3));
            fRec3[0] = ((0.999f * fRec3[1]) + fSlow10);
            fRec4[0] = ((0.70710677f * fTemp1) - ((0.70710677f * fTemp0) + (fSlow5 * ((fSlow6 * fRec4[1]) + (fSlow7 * fRec4[2])))));
            float fTemp5 = (fRec3[0] * ((fSlow5 * ((fSlow1 * (fRec4[2] + (fRec4[0] + (2.0f * fRec4[1])))) - (fRec4[0] + fRec4[2]))) - (fSlow9 * fRec4[1])));
            //*output++ = (audio_sample)(0.70710677f * (fTemp4 + fTemp5));
            //*output++ = (audio_sample)((0.70710677f * fTemp4) - (0.70710677f * fTemp5));
            *output++ = (audio_sample)((0.70710677f * fTemp4) - (0.70710677f * fTemp5)); // channels were reversed originally
            *output++ = (audio_sample)(0.70710677f * (fTemp4 + fTemp5));
            fVec0[(IOTA & 1023)] = ((fSlow3 * (cosf(fRec0[0]) * fTemp2)) + (cosf(fRec2[0]) * fTemp3));
            *output++ = (audio_sample)fVec0[((IOTA - iSlow11) & 1023)];
            fRec0[1] = fRec0[0];
            fRec1[2] = fRec1[1];
            fRec1[1] = fRec1[0];
            fRec2[1] = fRec2[0];
            fRec3[1] = fRec3[0];
            fRec4[2] = fRec4[1];
            fRec4[1] = fRec4[0];
            IOTA++;
        }
    }

public:
    trifield_dsp(const dsp_preset &in) : fSamplingFreq(0)
    {
        parse_preset(cdx, phihf, philf, width, xover, in);
    }

    ~trifield_dsp() {}

    static void defaults(float &cdx, float &phihf, float &philf, float &width, unsigned int &xover)
    {
        cdx   = 0.0f;
        phihf = 55.7f;
        philf = 35.26f;
        width = 1.0f;
        xover = 3000;
    }

    static GUID g_get_guid()
    {
        // {3F2ED403-5895-4B71-9AF0-DFE5CC7921D3}
        static const GUID guid = { 0x3f2ed403, 0x5895, 0x4b71, { 0x9a, 0xf0, 0xdf, 0xe5, 0xcc, 0x79, 0x21, 0xd3 } };
        return guid;
    }

    static void g_get_name(pfc::string_base &p_out)
    {
        p_out = "Trifield LR→LRC Decoder";
    }

    virtual bool on_chunk(audio_chunk *chunk, abort_callback &p_abort)
    {
        if (chunk->get_channel_count() == 2) { // only operate on stereo signal
            if (chunk->get_sample_rate() != fSamplingFreq) {
                init(chunk->get_sample_rate());
            }

            unsigned int samples = chunk->get_sample_count();
            audio_sample *src = (audio_sample *)chunk->get_data();
            audio_chunk *data = insert_chunk(samples * 3);
            if (data) {
                data->set_data_size(samples * 3);
                data->set_sample_rate(fSamplingFreq);
                data->set_channels(3, audio_chunk::channel_front_left|audio_chunk::channel_front_right|audio_chunk::channel_front_center);
                data->set_sample_count(samples);
                audio_sample *dst = data->get_data();
                compute(samples, src, dst);
                return false;
            }
        } else {
            fSamplingFreq = 0;
        }

        return true;
    }

    virtual void flush()
    {
    }

    virtual double get_latency()
    {
        return 0.0;
    }

    virtual bool need_track_change_mark()
    {
        return false;
    }

    static bool g_get_default_preset(dsp_preset &p_out)
    {
        float c, h, l, w;
        unsigned int x;
        defaults(c, h, l, w, x);
        make_preset(c, h, l, w, x, p_out);
        return true;
    }

    static void g_show_config_popup(const dsp_preset &p_data, HWND p_parent, dsp_preset_edit_callback &p_callback)
    {
        ::RunDSPConfigPopup(p_data, p_parent, p_callback);
    }

    static bool g_have_config_popup()
    {
		return true;
    }

    static void make_preset(float cdx, float phihf, float philf, float width, unsigned int xover, dsp_preset &out)
    {
        dsp_preset_builder builder;
        builder << cdx;
        builder << phihf;
        builder << philf;
        builder << width;
        builder << xover;
        builder.finish(g_get_guid(), out);
    }

    static void parse_preset(float &cdx, float &phihf, float &philf, float &width, unsigned int &xover, const dsp_preset &in)
    {
        try {
            dsp_preset_parser parser(in);
            parser >> cdx;
            parser >> phihf;
            parser >> philf;
            parser >> width;
            parser >> xover;
        } catch(exception_io_data) {
            defaults(cdx, phihf, philf, width, xover);
        }
    }

    virtual void on_endoftrack(abort_callback &p_abort) {}
    virtual void on_endofplayback(abort_callback &p_abort) {}
};

static dsp_factory_t<trifield_dsp> foo_dsp_trifield;

class CMyDSPPopup : public CDialogImpl<CMyDSPPopup> {
public:
    CMyDSPPopup(const dsp_preset &initData, dsp_preset_edit_callback &callback) : m_initData(initData), m_callback(callback) {}

    enum { IDD = IDD_CONFIG };

    BEGIN_MSG_MAP(CMyDSPPopup)
    MSG_WM_INITDIALOG(OnInitDialog)
    COMMAND_HANDLER_EX(IDOK, BN_CLICKED, OnButton)
    COMMAND_HANDLER_EX(IDCANCEL, BN_CLICKED, OnButton)
    MSG_WM_HSCROLL(OnHScroll)
    COMMAND_HANDLER_EX(IDC_RESET, BN_CLICKED, OnButton)
    END_MSG_MAP()

private:
    void SetSliders()
    {
        m_slider_cdx.SetPos((int)(cdx / 0.005f + 0.5f));
        m_slider_phihf.SetPos((int)(phihf / 0.01f + 0.5f));
        m_slider_philf.SetPos((int)(philf / 0.01f + 0.5f));
        m_slider_width.SetPos((int)(width / 0.01f + 0.5f));
        m_slider_xover.SetPos((int)(xover / 50));
    }

    BOOL OnInitDialog(CWindow, LPARAM)
    {
        m_slider_cdx = GetDlgItem(IDC_CDX);
        m_slider_phihf = GetDlgItem(IDC_PHIHF);
        m_slider_philf = GetDlgItem(IDC_PHILF);
        m_slider_width = GetDlgItem(IDC_WIDTH);
        m_slider_xover = GetDlgItem(IDC_XOVER);
        m_slider_cdx.SetRange(0, 200);    // 0 - 1, step 0.005
        m_slider_phihf.SetRange(0, 9000); // 0 - 90, step 0.01
        m_slider_philf.SetRange(0, 9000); // 0 - 90, step 0.01
        m_slider_width.SetRange(0, 200);  // 0 - 2, step 0.01
        m_slider_xover.SetRange(7, 200);  // 350 - 10000, step 50
        m_slider_phihf.SetTicFreq(100);
        m_slider_philf.SetTicFreq(100);

        trifield_dsp::parse_preset(cdx, phihf, philf, width, xover, m_initData);

        if (cdx < 0.0f) cdx = 0.0f; else if (cdx > 1.0f) cdx = 1.0f;
        if (phihf < 0.0f) phihf = 0.0f; else if (phihf > 90.0f) phihf = 90.0f;
        if (philf < 0.0f) philf = 0.0f; else if (philf > 90.0f) philf = 90.0f;
        if (width < 0.0f) width = 0.0f; else if (width > 2.0f) width = 2.0f;
        if (xover < 350) xover = 350; else if (xover > 10000) xover = 10000;

        SetSliders();
        RefreshLabels();
        return TRUE;
    }

    void OnButton(UINT, int id, CWindow)
    {
        if (id == IDC_RESET) {
            trifield_dsp::defaults(cdx, phihf, philf, width, xover);
            SetSliders();
            Update();
        } else {
            EndDialog(id);
        }
    }

    void OnHScroll(UINT nSBCode, UINT nPos, CScrollBar pScrollBar)
    {
        Update();
    }

    void Update()
    {
        cdx = (float)m_slider_cdx.GetPos() * 0.005f;
        phihf = (float)m_slider_phihf.GetPos() * 0.01f;
        philf = (float)m_slider_philf.GetPos() * 0.01f;
        width = (float)m_slider_width.GetPos() * 0.01f;
        xover = m_slider_xover.GetPos() * 50;

        dsp_preset_impl preset;
        trifield_dsp::make_preset(cdx, phihf, philf, width, xover, preset);
        m_callback.on_preset_changed(preset);
        RefreshLabels();
    }

    void RefreshLabels()
    {
        char temp[32];
        unsigned int t = (unsigned int)(cdx * 1000 + 0.5f);
        wsprintfA(temp, "%u.%03u", t/1000, t%1000);
        ::uSetDlgItemText(*this, IDC_CDX_D, temp);
        t = (unsigned int)(phihf * 100 + 0.5f);
        wsprintfA(temp, "%u.%02u", t/100, t%100);
        ::uSetDlgItemText(*this, IDC_PHIHF_D, temp);
        t = (unsigned int)(philf * 100 + 0.5f);
        wsprintfA(temp, "%u.%02u", t/100, t%100);
        ::uSetDlgItemText(*this, IDC_PHILF_D, temp);
        t = (unsigned int)(width * 100 + 0.5f);
        wsprintfA(temp, "%u.%02u", t/100, t%100);
        ::uSetDlgItemText(*this, IDC_WIDTH_D, temp);
        wsprintfA(temp, "%u", xover);
        ::uSetDlgItemText(*this, IDC_XOVER_D, temp);
    }

    const dsp_preset &m_initData;
    dsp_preset_edit_callback &m_callback;
    CTrackBarCtrl m_slider_cdx;
    CTrackBarCtrl m_slider_phihf;
    CTrackBarCtrl m_slider_philf;
    CTrackBarCtrl m_slider_width;
    CTrackBarCtrl m_slider_xover;
    float cdx, phihf, philf, width;
    unsigned int xover;
};

static void RunDSPConfigPopup(const dsp_preset &p_data, HWND p_parent, dsp_preset_edit_callback &p_callback)
{
    CMyDSPPopup popup(p_data, p_callback);
    if (popup.DoModal(p_parent) != IDOK) p_callback.on_preset_changed(p_data);
}
